//--
//--	animwind.cpp
//--
//--	TBitmapAnimatorWindow Class Module
//--
//--	  This class keeps its own private DC set up for fast blting from the
//--	WinGDC.  It also manages the offscreen buffer with simple DirtyRect
//--	processing and virtual function calls for buffer/screen refresh.
//--	You can draw to this buffer using the normal TDC functions or directly
//--	write bits via the "pOffScreenBits" pointer (the big advantage of
//--	WinG!).
//--	  Derived classes can simply set "DirtyRect" to the area of the
//--	offscreen buffer that needs refreshing and call the RefreshScreen()
//-- 	function.  RefreshScreen() will make virtual calls to DrawBackground()
//--	and DrawForeground() to let the derived class(es) refresh the offscreen
//--	buffer and then will Blt the DirtyRect portion to the window (screen).
//--
//--


#include "samppch.h"
#include <owl\window.h>
#include <owl\module.h>
#include <owl\point.h>
#include <owl\dc.h>
#include <owl\gdiobjec.h>
#include "sampdefs.h"
#include "sample.rh"
#include "wingdc.h"
#include "palette.h"
#include "animwind.h"


DEFINE_RESPONSE_TABLE1(TBitmapAnimatorWindow, TWindow)
	EV_WM_SIZE,
END_RESPONSE_TABLE;


//--------------------------------------------------
//  TBitmapAnimatorWindow::TBitmapAnimatorWindow()
//--------------------------------------------------
TBitmapAnimatorWindow::TBitmapAnimatorWindow(TWindow *pParWin,
											 const char far *pszTitle,
											 TModule *pMod) :
									 TWindow(pParWin,pszTitle,pMod)
{
	nWidth = nHeight = 0;
	pClientDC = 0;  // Allocated in SetupWindow()
	pBufferDC = 0;  // Allocated in SetupWindow()
	pPalette = new TIdentityPalette();
	hbmOldMonoBitmap = 0;
	pOffScreenBits = 0;
	lBMPOrientation = 1L;
	DirtyRect.SetNull();
}


//---------------------------------------------------
//  TBitmapAnimatorWindow::~TBitmapAnimatorWindow()
//---------------------------------------------------
TBitmapAnimatorWindow::~TBitmapAnimatorWindow()
{
	delete pPalette;
	delete pClientDC;  // Usually deleted in CleanupWindow()
	delete pBufferDC;  // Usually deleted in CleanupWindow()

	Destroy();
}


//-------------------------------------------
//  TBitmapAnimatorWindow::GetWindowClass()
//-------------------------------------------
void TBitmapAnimatorWindow::GetWindowClass(WNDCLASS &WndClass)
{
	TWindow::GetWindowClass(WndClass);

	WndClass.style = CS_OWNDC;
	WndClass.hbrBackground = (HBRUSH)GetStockObject(NULL_BRUSH);
}


//----------------------------------------
//  TBitmapAnimatorWindow::SetupWindow()
//----------------------------------------
void TBitmapAnimatorWindow::SetupWindow()
{
	// Pass to base class
	TWindow::SetupWindow();

	// Let derived classes fill the identity palette
	InitPalette();

	// Setup our owned DC
	pClientDC = new TClientDC(HWindow);
	pClientDC->SetStretchBltMode(COLORONCOLOR);
	pClientDC->SelectObject(*pPalette);
	pClientDC->RealizePalette();

	// Init WinG stuff
	// Allocate a TWinGDC
	try
	{
		pBufferDC = new TWinGDC();
	}
	catch (...)
	{
		pBufferDC = 0;
		::PostQuitMessage(1);
		return;
	}

	// Build and select an off-screen DIB for the WinGDC
	//  (Start with a dummy size just so we can save the old selected
	//    bitmap...the WM_SIZE handler will create the actual size
	//    offscreen bitmap)
	HBITMAP hBM = BuildOffScreenDIB(32,32);
	hbmOldMonoBitmap = (HBITMAP)::SelectObject((HDC)*pBufferDC,hBM);
}


//------------------------------------------
//  TBitmapAnimatorWindow::CleanupWindow()
//------------------------------------------
void TBitmapAnimatorWindow::CleanupWindow()
{
	// Select orig bitmap back in to the WinGDC and delete the current
	//    off-screen bitmap
	if (pBufferDC)
	{
		HBITMAP hBM = (HBITMAP)::SelectObject((HDC)*pBufferDC,
														   hbmOldMonoBitmap);
		::DeleteObject(hBM);
	}

	// Delete the WinG DC
	delete pBufferDC;
	pBufferDC = 0;

	// Cleanup our owned DC
	if (pClientDC)  pClientDC->RestoreObjects();
	delete pClientDC;
	pClientDC = 0;

	// Pass to base class
	TWindow::CleanupWindow();
}


//----------------------------------------------
//  TBitmapAnimatorWindow::BuildOffScreenDIB()
//----------------------------------------------
HBITMAP TBitmapAnimatorWindow::BuildOffScreenDIB(int nXSize, int nYSize)
{
	if (!pBufferDC)  return 0;

	// Query WinG for fastest DIB format - force to 8-bit/no compression
	if (pBufferDC->WinGRecommendDIBFormat((BITMAPINFO far *)&BufferHeader))
	{
		BufferHeader.Header.biBitCount = 8;
		BufferHeader.Header.biCompression = BI_RGB;
		lBMPOrientation = BufferHeader.Header.biHeight;
	}
	// If WinGRecommendDIBFormat() failed we'll just build our own 8-bit DIB
	else
	{
		BufferHeader.Header.biSize = sizeof(BITMAPINFOHEADER);
		BufferHeader.Header.biPlanes = 1;
		BufferHeader.Header.biBitCount = 8;
		BufferHeader.Header.biCompression = BI_RGB;
		BufferHeader.Header.biSizeImage = 0;
		BufferHeader.Header.biClrUsed = 0;
		BufferHeader.Header.biClrImportant = 0;
	}

	BufferHeader.Header.biWidth = nXSize;
	BufferHeader.Header.biHeight = nYSize * lBMPOrientation;

	// Set up the offscreen DIBs color table from the colors in pPalette
	//   Allocate and init a temporary PALETTEENTRY array
	uint nNumEntries = pPalette->GetNumEntries();
	PALETTEENTRY *pPalEntries = new PALETTEENTRY[nNumEntries];
	//   Fill the array with the palette entries
	::GetPaletteEntries((HPALETTE)*pPalette,0,nNumEntries,pPalEntries);
	//   Copy identity palette entries to our off-screen DIB header
	for (uint nCounter = 0; nCounter < nNumEntries; ++nCounter)
	{
		BufferHeader.aColors[nCounter].rgbRed =
									  pPalEntries[nCounter].peRed;
		BufferHeader.aColors[nCounter].rgbGreen =
									  pPalEntries[nCounter].peGreen;
		BufferHeader.aColors[nCounter].rgbBlue =
									  pPalEntries[nCounter].peBlue;
		BufferHeader.aColors[nCounter].rgbReserved = 0;
	}
	//   Free the temporary PALETTEENTRY array
	delete pPalEntries;

	// Create the off-screen DIB
	return (pBufferDC->WinGCreateBitmap((BITMAPINFO far *)&BufferHeader.Header,
														   &pOffScreenBits));
}


//------------------------------------------------
//  TBitmapAnimatorWindow::RebuildOffScreenDIB()
//------------------------------------------------
void TBitmapAnimatorWindow::RebuildOffScreenDIB(int nXSize, int nYSize)
{
	if(pBufferDC)
	{
		HBITMAP hBM;

		if (!nXSize)  nXSize = 32;
		if (!nYSize)  nYSize = 32;

		BufferHeader.Header.biWidth = nXSize;
		BufferHeader.Header.biHeight = nYSize * lBMPOrientation;
		hBM = pBufferDC->WinGCreateBitmap((BITMAPINFO far *)
											   &BufferHeader.Header,
															&pOffScreenBits);

		//  Select it in and delete the old one
		hBM = (HBITMAP)::SelectObject((HDC)*pBufferDC,hBM);
		::DeleteObject(hBM);
	}
}


//-------------------------------------------
//  TBitmapAnimatorWindow::DrawForeground()
//-------------------------------------------
void TBitmapAnimatorWindow::DrawForeground(TRect &/*UpdateRect*/)
{
	//** Placeholder
}


//-------------------------------------------
//  TBitmapAnimatorWindow::DrawBackground()
//-------------------------------------------
void TBitmapAnimatorWindow::DrawBackground(TRect &UpdateRect)
{
	pBufferDC->PatBlt(UpdateRect.left,UpdateRect.top,
						   UpdateRect.Width(),UpdateRect.Height(),BLACKNESS);
}


//------------------------------------------
//  TBitmapAnimatorWindow::RefreshScreen()
//------------------------------------------
void TBitmapAnimatorWindow::RefreshScreen()
{
	if (pBufferDC && !DirtyRect.IsEmpty())
	{
		DrawBackground(DirtyRect);
		DrawForeground(DirtyRect);

		pBufferDC->WinGBitBlt((HDC)*pClientDC,
								DirtyRect.left,
								DirtyRect.top,
								DirtyRect.Width(),
								DirtyRect.Height(),
								DirtyRect.left,
								DirtyRect.top);

		DirtyRect.SetNull();
	}
}


//----------------------------------
//  TBitmapAnimatorWindow::Paint()
//----------------------------------
void TBitmapAnimatorWindow::Paint(TDC &dc, bool, TRect &UpdateRect)
{
	dc.SelectObject(*pPalette);
	dc.RealizePalette();

	DirtyRect = UpdateRect;

	RefreshScreen();
}


//-----------------------------------
//  TBitmapAnimatorWindow::EvSize()
//-----------------------------------
void TBitmapAnimatorWindow::EvSize(uint /*nType*/, TSize &Size)
{
	//
	// Keep the dimensions of the offscreen buffer equal to the dimensions
	//  of the window
	//

	if (pBufferDC && Size.cx && Size.cy)  // During create, gets called with 0s
	{
		nWidth = Size.cx;
		nHeight = Size.cy;

		// resize off-screen DIB
		RebuildOffScreenDIB(nWidth,nHeight);

		// force redraw of window
		Invalidate();
	}
	else
	{
		DefaultProcessing();
	}
}


//---------------------------------------------
//  TBitmapAnimatorWindow::EvPaletteChanged()
//---------------------------------------------
void TBitmapAnimatorWindow::EvPaletteChanged(HWND hwnd)
{
	if (hwnd != HWindow)  EvQueryNewPalette();
}


//----------------------------------------------
//  TBitmapAnimatorWindow::EvQueryNewPalette()
//----------------------------------------------
bool TBitmapAnimatorWindow::EvQueryNewPalette()
{
	pClientDC->SelectObject(*pPalette);

	int nClrChanged = pClientDC->RealizePalette();

	if (nClrChanged)  Invalidate();

	return (nClrChanged != 0);
}



